from .hash import ckb_hash
from .molecule import extend_uint32, extend_uint64, extend_uint128, extend_bytes_array
from .types import Header, HexH256


def extend_serialized_raw_header(buffer: bytearray, header: Header) -> bytearray:
    # struct RawHeader
    # {
    #     version: Uint32,
    #     compact_target: Uint32,
    #     timestamp: Uint64,
    #     number: Uint64,
    #     epoch: Uint64,
    #     parent_hash: Byte32,
    #     transactions_root: Byte32,
    #     proposals_hash: Byte32,
    #     uncles_hash: Byte32,
    #     dao: Byte32,
    # }
    extend_uint32(buffer, header['version'])
    extend_uint32(buffer, header['compact_target'])
    extend_uint64(buffer, header['timestamp'])
    extend_uint64(buffer, header['number'])
    extend_uint64(buffer, header['epoch'])
    extend_bytes_array(buffer, header['parent_hash'])
    extend_bytes_array(buffer, header['transactions_root'])
    extend_bytes_array(buffer, header['proposals_hash'])
    extend_bytes_array(buffer, header['uncles_hash'])
    extend_bytes_array(buffer, header['dao'])

    return buffer


def extend_serialized_header(buffer: bytearray, header: Header) -> bytearray:
    # struct Header
    # {
    #     raw: RawHeader,
    #     nonce: Uint128,
    # }
    extend_serialized_raw_header(buffer, header)
    extend_uint128(buffer, header['nonce'])

    return buffer


def header_hash(header: Header) -> HexH256:
    buffer = bytearray()
    extend_serialized_header(buffer, header)
    return ckb_hash(buffer)


def pow_hash(header: Header) -> HexH256:
    buffer = bytearray()
    extend_serialized_raw_header(buffer, header)
    return ckb_hash(buffer)



from typing import Sequence

from . import uncle_block, transaction, header, hex_coder
from .cbmt import cbmt_root
from .hash import ckb_hasher, H256_ZEROS
from .transaction import transaction_hash, transaction_witness_hash
from .types import HexH256, ProposalShortId, Block, Header, Transaction, UncleBlock, BlockTemplate


def transactions_root(transactions: Sequence[Transaction]) -> HexH256:
    transaction_hash_root = cbmt_root([transaction_hash(tx) for tx in transactions])
    witness_hash_root = cbmt_root([transaction_witness_hash(tx) for tx in transactions])

    hasher = ckb_hasher()
    hasher.update(hex_coder.hex_to_bytes(transaction_hash_root))
    hasher.update(hex_coder.hex_to_bytes(witness_hash_root))

    return '0x' + hasher.hexdigest()


def proposals_hash(proposals: Sequence[ProposalShortId]) -> HexH256:
    if len(proposals) == 0:
        return H256_ZEROS

    hasher = ckb_hasher()

    for id in proposals:
        hasher.update(hex_coder.hex_to_bytes(id))

    return '0x' + hasher.hexdigest()


def uncles_hash(uncles: Sequence[UncleBlock]) -> HexH256:
    if len(uncles) == 0:
        return H256_ZEROS

    hasher = ckb_hasher()

    for uncle in uncles:
        hasher.update(hex_coder.hex_to_bytes(header.header_hash(uncle['header'])))

    return '0x' + hasher.hexdigest()


def from_template(template: BlockTemplate) -> Block:
    transactions = [transaction.from_cellbase_template(template['cellbase'])]
    transactions.extend(transaction.from_template(tx) for tx in template['transactions'])

    proposals = template['proposals']

    uncles = list(uncle_block.from_template(uncle) for uncle in template['uncles'])

    header = Header(
        version=template['version'],
        compact_target=template['compact_target'],
        timestamp=template['current_time'],
        number=template['number'],
        epoch=template['epoch'],
        transactions_root=transactions_root(transactions),
        proposals_hash=proposals_hash(proposals),
        uncles_hash=uncles_hash(uncles),
        parent_hash=template['parent_hash'],
        dao=template['dao'],
        nonce='0x0'
    )

    return Block(
        header=header,
        uncles=uncles,
        transactions=transactions,
        proposals=proposals
    )



from typing import Optional, Sequence, NewType, Union

from typing_extensions import TypedDict

HexBytes = NewType('HexBytes', str)
HexByte32 = NewType('HexByte32', HexBytes)
HexH256 = NewType('HexH256', HexBytes)
HexInt = NewType('HexInt', str)

BlockNumber = NewType('BlockNumber', HexInt)
EpochNumber = NewType('EpochNumber', HexInt)
EpochNumberWithFraction = NewType('EpochNumberWithFraction', HexInt)
ProposalShortId = NewType('ProposalShortId', HexBytes)

DepType = Union['code', 'dep_group']
ScriptHashType = Union['data', 'type']

Header = TypedDict('Header', {
    'version': HexInt,
    'compact_target': HexInt,
    'timestamp': HexInt,
    'number': BlockNumber,
    'epoch': EpochNumberWithFraction,
    'parent_hash': HexH256,
    'transactions_root': HexH256,
    'proposals_hash': HexH256,
    'uncles_hash': HexH256,
    'dao': HexByte32,
    'nonce': HexInt,
    'hash': Optional[HexH256]
})

UncleBlock = TypedDict('UncleBlock', {
    'header': Header,
    'proposals': Sequence[ProposalShortId]
})

OutPoint = TypedDict('OutPoint', {
    'tx_hash': HexH256,
    'index': HexInt
})

CellDep = TypedDict('CellDep', {
    'out_point': OutPoint,
    'dep_type': DepType
})

CellInput = TypedDict('CellInput', {
    'since': HexInt,
    'previous_output': OutPoint
})

Script = TypedDict('Script', {
    'code_hash': HexH256,
    'hash_type': ScriptHashType,
    'args': HexBytes
})

CellOutput = TypedDict('CellOutput', {
    'capacity': HexInt,
    'lock': Script,
    'type': Optional[Script]
})

Transaction = TypedDict('Transaction', {
    'version': HexInt,
    'cell_deps': Sequence[CellDep],
    'header_deps': Sequence[HexH256],
    'inputs': Sequence[CellInput],
    'outputs': Sequence[CellOutput],
    'outputs_data': Sequence[HexBytes],
    'witnesses': Sequence[HexBytes],
    'hash': Optional[HexH256]
})

Block = TypedDict('Block', {
    'header': Header,
    'uncles': Sequence[UncleBlock],
    'transactions': Sequence[Transaction],
    'proposals': Sequence[ProposalShortId]
})

Epoch = TypedDict('Epoch', {
    'number': EpochNumber,
    'start_number': BlockNumber,
    'length': BlockNumber,
    'compact_target': HexInt
})

BlockReward = TypedDict('BlockReward', {
    'total': HexInt,
    'primary': HexInt,
    'secondary': HexInt,
    'tx_fee': HexInt,
    'proposal_reward': HexInt
})

UncleTemplate = TypedDict('UncleTemplate', {
    'hash': HexH256,
    'required': bool,
    'proposals': Sequence[ProposalShortId],
    'header': Header
})

TransactionTemplate = TypedDict('TransactionTemplate', {
    'hash': HexH256,
    'required': bool,
    'cycles': Optional[HexInt],
    'depends': Optional[Sequence[HexInt]],
    'data': Transaction
})

CellbaseTemplate = TypedDict('CellbaseTemplate', {
    'hash': HexH256,
    'cycles': Optional[HexInt],
    'data': Transaction
})

BlockTemplate = TypedDict('BlockTemplate', {
    'version': HexInt,
    'compact_target': HexInt,
    'current_time': HexInt,
    'number': BlockNumber,
    'epoch': EpochNumberWithFraction,
    'parent_hash': HexH256,
    'cycles_limit': HexInt,
    'bytes_limit': HexInt,
    'uncles_count_limit': HexInt,
    'uncles': Sequence[UncleTemplate],
    'transactions': Sequence[TransactionTemplate],
    'proposals': Sequence[ProposalShortId],
    'cellbase': CellbaseTemplate,
    'work_id': HexInt,
    'dao': HexByte32
})

WitnessArgs = TypedDict('WitnessArgs', {
    'lock': HexBytes,
    'input_type': HexBytes,
    'output_type': HexBytes,
})



from .types import Transaction, HexBytes
from .transaction import transaction_hash, extend_serialized_raw_transaction
from .hash import ckb_hasher
from .hex_coder import hex_to_bytes, hex_from_bytes
from .molecule import extend_uint64
from coincurve import PrivateKey
from .witness_args import new_witness_args, extend_serialized_witness_args

def sign_tx(tx: Transaction, private_key: HexBytes) -> Transaction:
    tx_hash = transaction_hash(tx)
    hasher = ckb_hasher()
    hasher.update(hex_to_bytes(tx_hash))

    witness_args = new_witness_args()
    witness_buffer = bytearray()
    extend_serialized_witness_args(witness_buffer, witness_args)
    len_buffer = bytearray()
    extend_uint64(len_buffer, str(len(witness_buffer)))
    hasher.update(len_buffer)
    hasher.update(witness_buffer)

    witnesses = tx['witnesses'][1:]
    for witness in witnesses:
        len_buffer = bytearray()
        extend_uint64(len_buffer, len(witness))
        hasher.update(len_buffer)
        hasher.update(hex_to_bytes(witnesses))

    message = '0x' + hasher.hexdigest()

    _private_key = PrivateKey.from_hex(private_key)
    signature = _private_key.sign_recoverable(hex_to_bytes(message), None)

    witness_args = new_witness_args(lock=hex_from_bytes(signature))
    witness_buffer = bytearray()
    serialized_witness_args = extend_serialized_witness_args(
        witness_buffer, witness_args)

    tx['witnesses'][0] = hex_from_bytes(witness_buffer)

    return tx



from unittest import TestCase

from . import cbmt
from .hash import H256_ZEROS
from .hex_coder import hex_from_bytes, hex_to_bytes


def merge(a, b):
    return hex_from_bytes(cbmt.merge(hex_to_bytes(a), hex_to_bytes(b)))


TEST_ITEMS = [
    '0x2751b0faa8b0b63ffb74e0de06b05370c583e9f0befabdde3dc03332fbb58f57',
    '0x88ac55ef5d947349b937cb3f307dd7b1650847f5389dcc1ff999df34c61ee9f9',
    '0xfc337768c8817bd291eaa770550b07c28c3aec01d0046ac654a8098ee9c433a5',
    '0xb92f8fa579cf33d1085b805f6ec877c6600b7f7a5ac3433aedec87a31f69dfa1',
    '0xea29021190bc3891016d1a69b3d7ea44a33b62ebe9717c458b98c7b5e4227d8f',
    '0xc7b99222cb6647bcb28af98f70c6417b73e4701088ab82cc988f857b880ca055',
    '0x4c493215c235f3546b66aaec1eda68044c4ab0ada549433356789596108a7cda',
    '0xdf2597213b983c0d4f7fd426e38128bb609ff58e0b05c6e8f2c7c3ea041e623d',
    '0x65b6a53f6aa7413ece5b100e2e2c203bef6613262815ea621348427a9941d884'
]


class TestCbmt(TestCase):
    def test_empty_tree(self):
        self.assertEqual(cbmt.cbmt_root([]), H256_ZEROS)

    def test_single_item(self):
        self.assertEqual(cbmt.cbmt_root(TEST_ITEMS[0:1]), TEST_ITEMS[0])

    def test_two_items(self):
        self.assertEqual(cbmt.cbmt_root(TEST_ITEMS[0:2]), merge(TEST_ITEMS[0], TEST_ITEMS[1]))

    def test_three_items(self):
        self.assertEqual(cbmt.cbmt_root(TEST_ITEMS[0:3]), merge(merge(TEST_ITEMS[1], TEST_ITEMS[2]), TEST_ITEMS[0]))

    def test_four_items(self):
        self.assertEqual(cbmt.cbmt_root(TEST_ITEMS[0:4]),
                         merge(merge(TEST_ITEMS[0], TEST_ITEMS[1]), merge(TEST_ITEMS[2], TEST_ITEMS[3])))

    def test_five_items(self):
        self.assertEqual(cbmt.cbmt_root(TEST_ITEMS[0:5]),
                         merge(merge(merge(TEST_ITEMS[3], TEST_ITEMS[4]), TEST_ITEMS[0]),
                               merge(TEST_ITEMS[1], TEST_ITEMS[2])))

    def test_six_items(self):
        self.assertEqual(cbmt.cbmt_root(TEST_ITEMS[0:6]),
                         merge(merge(merge(TEST_ITEMS[2], TEST_ITEMS[3]), merge(TEST_ITEMS[4], TEST_ITEMS[5])),
                               merge(TEST_ITEMS[0], TEST_ITEMS[1])))

    def test_seven_items(self):
        self.assertEqual(cbmt.cbmt_root(TEST_ITEMS[0:7]),
                         merge(merge(merge(TEST_ITEMS[1], TEST_ITEMS[2]), merge(TEST_ITEMS[3], TEST_ITEMS[4])),
                               merge(merge(TEST_ITEMS[5], TEST_ITEMS[6]), TEST_ITEMS[0])))

    def test_eight_items(self):
        self.assertEqual(cbmt.cbmt_root(TEST_ITEMS[0:8]),
                         merge(merge(merge(TEST_ITEMS[0], TEST_ITEMS[1]), merge(TEST_ITEMS[2], TEST_ITEMS[3])),
                               merge(merge(TEST_ITEMS[4], TEST_ITEMS[5]), merge(TEST_ITEMS[6], TEST_ITEMS[7]))))

    def test_nine_items(self):
        self.assertEqual(cbmt.cbmt_root(TEST_ITEMS[0:9]),
                         merge(merge(merge(merge(TEST_ITEMS[7], TEST_ITEMS[8]), TEST_ITEMS[0]),
                                     merge(TEST_ITEMS[1], TEST_ITEMS[2])),
                               merge(merge(TEST_ITEMS[3], TEST_ITEMS[4]), merge(TEST_ITEMS[5], TEST_ITEMS[6]))))



from .types import UncleBlock, UncleTemplate


def from_template(template: UncleTemplate) -> UncleBlock:
    return UncleBlock(
        header=template['header'],
        proposals=template['proposals']
    )



from .types import HexBytes, HexInt


def hex_to_int(i: HexInt) -> int:
    return int(i, 0)


def hex_from_int(i: int) -> HexInt:
    return HexInt(hex(i))


def hex_to_bytes(b: HexBytes) -> bytes:
    return bytes.fromhex(b[2:])


def hex_from_bytes(b: bytes) -> HexBytes:
    return HexBytes('0x' + b.hex())



"""
 https://github.com/CipherWang/pyckbwallet/blob/master/address.py
"""

from . import segwit_addr as sa
from ..core.types import Script

FORMAT_TYPE_SHORT     = 0x01
FORMAT_TYPE_FULL_DATA = 0x02
FORMAT_TYPE_FULL_TYPE = 0x04

CODE_INDEX_SECP256K1_SINGLE = 0x00
CODE_INDEX_SECP256K1_MULTI  = 0x01

def generateShortAddress(args, code_index = CODE_INDEX_SECP256K1_SINGLE, network = "mainnet"):
    """ generate a short ckb address """
    hrp = {"mainnet": "ckb", "testnet": "ckt"}[network]
    hrpexp =  sa.bech32_hrp_expand(hrp)
    format_type  = FORMAT_TYPE_SHORT
    payload = bytes([format_type, code_index]) + bytes.fromhex(args)
    data_part = sa.convertbits(payload, 8, 5)
    values = hrpexp + data_part
    polymod = sa.bech32_polymod(values + [0, 0, 0, 0, 0, 0]) ^ 1
    checksum = [(polymod >> 5 * (5 - i)) & 31 for i in range(6)]
    combined = data_part + checksum
    addr = hrp + '1' + ''.join([sa.CHARSET[d] for d in combined])
    return addr

def generateFullAddress(script: Script, network = "mainnet"):
    format_type = {"data" : bytes([FORMAT_TYPE_FULL_DATA]),
                 "type" : bytes([FORMAT_TYPE_FULL_TYPE])}[script['hash_type']]
    hrp = {"mainnet": "ckb", "testnet": "ckt"}[network]
    hrpexp =  sa.bech32_hrp_expand(hrp)
    payload = bytes(format_type) + bytes.fromhex(script['code_hash'])
    payload += bytes.fromhex(script['args'])
    data_part = sa.convertbits(payload, 8, 5)
    values = hrpexp + data_part
    polymod = sa.bech32_polymod(values + [0, 0, 0, 0, 0, 0]) ^ 1
    checksum = [(polymod >> 5 * (5 - i)) & 31 for i in range(6)]
    combined = data_part + checksum
    addr = hrp + '1' + ''.join([sa.CHARSET[d] for d in combined])
    return addr


def decodeAddress(addr, network = "mainnet"):
    hrp = {"mainnet": "ckb", "testnet": "ckt"}[network]
    hrpgot, data = sa.bech32_decode(addr)
    if hrpgot != hrp or data == None:
        return False
    decoded = sa.convertbits(data, 5, 8, False)
    if decoded == None:
        return False
    payload = bytes(decoded)
    format_type = payload[0]
    if format_type == FORMAT_TYPE_SHORT:
        code_index = payload[1]
        pk = payload[2:].hex()
        return ("short", code_index, pk)
    elif format_type == FORMAT_TYPE_FULL_DATA or format_type == FORMAT_TYPE_FULL_TYPE:
        full_type = {FORMAT_TYPE_FULL_DATA:"Data", FORMAT_TYPE_FULL_TYPE:"Type"}[format_type]
        ptr = 1
        code_hash = payload[ptr : ptr+32].hex()
        ptr += 32
        args = payload[ptr :].hex()
        return ("full", full_type, code_hash, args)


import os
from typing import Optional

import jsonrpcclient

from .types import BlockNumber, EpochNumber, HexH256, HexInt, Header, Block, Transaction, Epoch, BlockReward, \
    BlockTemplate

DEFAULT_ENDPOINT = os.environ.get('CKB_RPC_URL', 'http://127.0.0.1:8114')


class RPCClient:
    def __init__(self, endpoint: str = DEFAULT_ENDPOINT):
        self.endpoint = endpoint

    def request(self, method: str, *args):
        return jsonrpcclient.request(self.endpoint, method, *args).data.result

    def get_block(self, hash: HexH256) -> Block:
        return self.request('get_block', hash)

    def get_block_by_number(self, number: BlockNumber) -> Block:
        return self.request('get_block_by_number', number)

    def get_header(self, hash: HexH256) -> Header:
        return self.request('get_header', hash)

    def get_header_by_number(self, number: BlockNumber) -> Header:
        return self.request('get_header_by_number', number)

    def get_transaction(self, hash: HexH256) -> Transaction:
        return self.request('get_transaction', hash)

    def get_block_hash(self, number: BlockNumber) -> HexH256:
        return self.request('get_block_hash', number)

    def get_tip_header(self) -> Header:
        return self.request('get_tip_header')

    def get_tip_block_number(self) -> BlockNumber:
        return self.request('get_tip_block_number')

    def get_current_epoch(self) -> Epoch:
        return self.request('get_current_epoch')

    def get_epoch_by_number(self, number: EpochNumber) -> Epoch:
        return self.request('get_epoch_by_number', number)

    def get_cellbase_output_capacity_details(self, hash: HexH256) -> BlockReward:
        return self.request('get_cellbase_output_capacity_details', hash)

    def get_block_template(self,
                           bytes_limit: Optional[HexInt] = None,
                           proposals_limit: Optional[HexInt] = None,
                           max_version: Optional[HexInt] = None
                           ) -> BlockTemplate:
        return self.request('get_block_template', bytes_limit, proposals_limit, max_version)

    def submit_block(self, work_id: HexInt, block: Block) -> HexH256:
        return self.request('submit_block', block)


rpc = RPCClient()



from .hex_coder import hex_to_int, hex_to_bytes
from .types import HexInt, HexBytes, HexByte32


class HeaderBuilder:
    def __init__(self, buffer: bytearray, items_len: int):
        self.buffer = buffer
        self.header_pos = len(buffer)
        self.offset_pos = self.header_pos + 4
        self.body_pos = self.offset_pos + items_len * 4

        if items_len == 0:
            # total bytes
            buffer.extend((4).to_bytes(4, 'little'))
        else:
            # total bytes placeholder
            buffer.extend(b'\0' * 4)
            # first item offset
            buffer.extend((items_len * 4 + 4).to_bytes(4, 'little'))
            self.offset_pos += 4
            # remaining items_len - 1 offsets placeholders
            if items_len > 1:
                buffer.extend(b'\0' * ((items_len - 1) * 4))

    def finish_item(self):
        offset = (len(self.buffer) - self.header_pos).to_bytes(4, 'little')

        if self.offset_pos < self.body_pos:
            self.buffer[self.offset_pos:self.offset_pos + 4] = offset
            self.offset_pos += 4
        else:
            self.buffer[self.header_pos:self.header_pos + 4] = offset


def extend_uint32(buffer: bytearray, n: HexInt):
    buffer.extend(hex_to_int(n).to_bytes(4, 'little'))


def extend_uint64(buffer: bytearray, n: HexInt):
    buffer.extend(hex_to_int(n).to_bytes(8, 'little'))


def extend_uint128(buffer: bytearray, n: HexInt):
    buffer.extend(hex_to_int(n).to_bytes(16, 'little'))


def extend_bytes_array(buffer: bytearray, b: HexBytes):
    buffer.extend(hex_to_bytes(b))


def extend_bytes32(buffer: bytearray, b: HexByte32):
    if len(b) < 32:
        buffer.extend(hex_to_bytes(
            '0x0000000000000000000000000000000000000000000000000000000000000000'))
    else:
        buffer.extend(hex_to_bytes(b))


def extend_bytes_fixvec(buffer: bytearray, b: HexBytes):
    binary = hex_to_bytes(b)
    buffer.extend(len(binary).to_bytes(4, 'little'))
    buffer.extend(binary)



from unittest import TestCase

from .transaction import transaction_hash, transaction_witness_hash

TEST_TRANSACTION = {'cell_deps': [], 'hash': '0xa09fdc7b0751b8dd11947bbba229ad16d825dfad18d31bc662f048233e2251f3',
                    'header_deps': [], 'inputs': [{'previous_output': {'index': '0xffffffff',
                                                                       'tx_hash': '0x0000000000000000000000000000000000000000000000000000000000000000'},
                                                   'since': '0xb0e'}], 'outputs': [{'capacity': '0x19ff7d6f4e',
                                                                                    'lock': {
                                                                                        'args': '0xdde7801c073dfb3464c7b1f05b806bb2bbb84e99',
                                                                                        'code_hash': '0x9bd7e06f3ecf4be0f2fcd2188b23f1b9fcc88e5d4b65a8637b17723bbda3cce8',
                                                                                        'hash_type': 'type'},
                                                                                    'type': None}],
                    'outputs_data': ['0x'], 'version': '0x0', 'witnesses': [
        '0x5d0000000c00000055000000490000001000000030000000310000009bd7e06f3ecf4be0f2fcd2188b23f1b9fcc88e5d4b65a8637b17723bbda3cce80114000000dde7801c073dfb3464c7b1f05b806bb2bbb84e990400000000000000']}


class TransactionTest(TestCase):
    def test_transaction_hash(self):
        self.assertEqual(transaction_hash(TEST_TRANSACTION), TEST_TRANSACTION['hash'])

    def test_transaction_witness_hash(self):
        self.assertEqual(transaction_witness_hash(TEST_TRANSACTION),
                         '0xa94ece15a2c5495f0aaf6e3e2fc33661aa6822e6badbc2fbe75e8452a754061a')



from .types import HexBytes, WitnessArgs
from .molecule import HeaderBuilder, extend_uint32, extend_uint64, extend_bytes_array, extend_bytes_fixvec
from .hex_coder import hex_from_bytes, hex_to_bytes

SIGNATURE_PLACEHOLDER = "0x0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"


def new_witness_args(lock=SIGNATURE_PLACEHOLDER):
    return {'lock': lock, 'input_type': '', 'output_type': ''}


def extend_serialized_witness_args(buffer: bytearray, witness_args: WitnessArgs) -> bytearray:
    # table WitnessArgs
    # {
    #     lock: BytesVec,
    #     input_type: BytesVec,
    #     output_type: BytesVec,
    # }

    witness_args_header = HeaderBuilder(buffer, 3)

    extend_bytes_fixvec(buffer, witness_args['lock'])
    witness_args_header.finish_item()

    if witness_args['input_type'] != '' and witness_args['input_type'] != '0x':
        extend_bytes_fixvec(buffer, witness_args['input_type'])
    witness_args_header.finish_item()

    if witness_args['input_type'] != '' and witness_args['output_type'] != '0x':
        extend_bytes_fixvec(buffer, witness_args['output_type'])
    witness_args_header.finish_item()

    return buffer



import hashlib

from .types import HexH256

H256_ZEROS: HexH256 = '0x0000000000000000000000000000000000000000000000000000000000000000'


def ckb_hasher():
    return hashlib.blake2b(digest_size=32, person=b'ckb-default-hash')


def ckb_hash(message: bytes) -> HexH256:
    hasher = ckb_hasher()
    hasher.update(message)
    return '0x' + hasher.hexdigest()

def ckb_blake160(message: bytes) -> HexH256:
    return ckb_hash(message)[0:42]


from .hash import ckb_hash
from .molecule import HeaderBuilder, extend_uint32, extend_uint64, extend_bytes_array, extend_bytes_fixvec, extend_bytes32
from .types import Transaction, TransactionTemplate, CellbaseTemplate, CellOutput, Script, HexH256


def extend_serialized_script(buffer: bytearray, script: Script) -> bytearray:
    # table Script
    # {
    #     code_hash: Byte32,
    #     hash_type: byte,
    #     args: Bytes,
    # }
    header = HeaderBuilder(buffer, 3)

    extend_bytes_array(buffer, script['code_hash'])
    header.finish_item()

    buffer.append(0 if script['hash_type'] == 'data' else 1)
    header.finish_item()

    extend_bytes_fixvec(buffer, script['args'])
    header.finish_item()

    return buffer


def extend_serialized_cell_output(buffer: bytearray, cell_output: CellOutput) -> bytearray:
    # table CellOutput
    # {
    #     capacity: Uint64,
    #     lock: Script,
    #     type_: ScriptOpt,
    # }
    header = HeaderBuilder(buffer, 3)

    extend_uint64(buffer, cell_output['capacity'])
    header.finish_item()

    extend_serialized_script(buffer, cell_output['lock'])
    header.finish_item()

    if 'type' in cell_output and cell_output['type'] is not None:
        extend_serialized_script(buffer, cell_output['type'])
    header.finish_item()

    return buffer


def extend_serialized_raw_transaction(buffer: bytearray, transaction: Transaction) -> bytearray:
    # table RawTransaction
    # {
    #     version: Uint32,
    #     cell_deps: CellDepVec,
    #     header_deps: Byte32Vec,
    #     inputs: CellInputVec,
    #     outputs: CellOutputVec,
    #     outputs_data: BytesVec,
    # }
    transaction_header = HeaderBuilder(buffer, 6)

    extend_uint32(buffer, transaction['version'])
    transaction_header.finish_item()

    # struct OutPoint
    # {
    #     tx_hash: Byte32,
    #     index: Uint32,
    # }

    buffer.extend(len(transaction['cell_deps']).to_bytes(4, 'little'))
    for cell_dep in transaction['cell_deps']:
        # struct CellDep
        # {
        #     out_point: OutPoint,
        #     dep_type: byte,
        # }
        extend_bytes_array(buffer, cell_dep['out_point']['tx_hash'])
        extend_uint32(buffer, cell_dep['out_point']['index'])
        buffer.append(0 if cell_dep['dep_type'] == 'code' else 1)
    transaction_header.finish_item()

    buffer.extend(len(transaction['header_deps']).to_bytes(4, 'little'))
    for header_dep in transaction['header_deps']:
        extend_bytes32(buffer, header_dep)
    transaction_header.finish_item()

    buffer.extend(len(transaction['inputs']).to_bytes(4, 'little'))
    for input in transaction['inputs']:
        # struct CellInput
        # {
        #     since: Uint64,
        #     previous_output: OutPoint,
        # }
        extend_uint64(buffer, input['since'])
        extend_bytes_array(buffer, input['previous_output']['tx_hash'])
        extend_uint32(buffer, input['previous_output']['index'])
    transaction_header.finish_item()

    outputs_header = HeaderBuilder(buffer, len(transaction['outputs']))
    for output in transaction['outputs']:
        extend_serialized_cell_output(buffer, output)
        outputs_header.finish_item()
    transaction_header.finish_item()

    outputs_data_header = HeaderBuilder(
        buffer, len(transaction['outputs_data']))
    for output_data in transaction['outputs_data']:
        extend_bytes_fixvec(buffer, output_data)
        outputs_data_header.finish_item()
    transaction_header.finish_item()

    return buffer


def extend_serialized_transaction(buffer: bytearray, transaction: Transaction) -> bytearray:
    # table Transaction
    # {
    #     raw: RawTransaction,
    #     witnesses: BytesVec,
    # }
    transaction_header = HeaderBuilder(buffer, 2)

    extend_serialized_raw_transaction(buffer, transaction)
    transaction_header.finish_item()

    witnesses_header = HeaderBuilder(buffer, len(transaction['witnesses']))
    for witness in transaction['witnesses']:
        extend_bytes_fixvec(buffer, witness)
        witnesses_header.finish_item()
    transaction_header.finish_item()

    return buffer


def transaction_hash(transaction: Transaction) -> HexH256:
    buffer = bytearray()
    extend_serialized_raw_transaction(buffer, transaction)
    return ckb_hash(buffer)


def transaction_witness_hash(transaction: Transaction) -> HexH256:
    buffer = bytearray()
    extend_serialized_transaction(buffer, transaction)
    return ckb_hash(buffer)


def from_cellbase_template(template: CellbaseTemplate) -> Transaction:
    return template['data']


def from_template(template: TransactionTemplate) -> Transaction:
    return template['data']



import os
import io
from setuptools import setup, find_packages

HERE = os.path.dirname(os.path.realpath(__file__))

README = os.path.join(HERE, 'README.md')
with io.open(README, encoding='utf-8') as f:
    long_description = f.read()

VERSION = os.path.join(HERE, 'ckb_toolkit', 'version.py')
with io.open(VERSION, encoding='utf-8') as f:
    package = {}
    exec(f.read(), package)
    version = package['VERSION']

setup(name='ckb-toolkit',
      version=version,
      description='Nervos CKB Toolkit',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/duanyytop/ckb-python-toolkit',
      author='dylan',
      author_email='dylan@nervina.io',
      license='MIT',
      packages=find_packages(),
      install_requires=['jsonrpcclient[requests]',
                        'typing-extensions', 'coincurve'],
      scripts=[],
      zip_safe=False,
      classifiers=[
          'Development Status :: 2 - Pre-Alpha',
          'Environment :: Console',
      ],
      include_package_data=True,
      )



from typing import Sequence

from .hash import H256_ZEROS, ckb_hasher
from .hex_coder import hex_to_bytes, hex_from_bytes
from .types import HexH256


# Find max pow of two which <= n
def max_pow_of_two(n: int) -> int:
    r = 1
    while n > 0:
        n = n >> 1
        r = r << 1

    return r >> 1


def merge(left: bytes, right: bytes) -> bytes:
    h = ckb_hasher()
    h.update(left)
    h.update(right)
    return h.digest()


def cbmt_root(items: Sequence[HexH256]) -> HexH256:
    items_len = len(items)
    if items_len == 0:
        return H256_ZEROS

    nodes_len = max_pow_of_two(items_len)
    merge_len = items_len - nodes_len
    merge_offset = items_len - merge_len * 2
    nodes = []

    # fill the lowest complete layer
    for i in range(merge_len):
        nodes.append(merge(hex_to_bytes(items[merge_offset + i * 2]), hex_to_bytes(items[merge_offset + i * 2 + 1])))
    for i in range(nodes_len - merge_len):
        nodes.append(hex_to_bytes(items[i]))

    remaining_nodes_len = nodes_len
    while remaining_nodes_len > 1:
        remaining_nodes_len = remaining_nodes_len >> 1
        for i in range(remaining_nodes_len):
            nodes[i] = merge(nodes[i * 2], nodes[i * 2 + 1])

    return HexH256(hex_from_bytes(nodes[0]))
